from typing import Any, Dict, Callable, Optional

def dict_traverser(
	data: Any,
	handlers: Dict[Any, Callable] = {},
	key: Optional[str] = None,
):
	for type, handler in handlers.items():
		if isinstance(data, type):
			new_res = handler(key, data)
			return new_res
	
	if isinstance(data, list) or isinstance(data, tuple):
		if isinstance(data, list):
			return [dict_traverser(item, handlers, key) for item in data]
		return (dict_traverser(item, handlers, key) for item in data)
	if not isinstance(data, dict):
		return data
	
	new_res = {}
	for k, v in data.items():
		new_res[k] = dict_traverser(v, handlers, k)
	return new_res



import os
import time
from pathlib import Path

import pytest
from bohrium_open_sdk import OpenSDK, UploadInputItem
from dotenv import find_dotenv, load_dotenv
from glom import glom

load_dotenv(find_dotenv(), override=True)


def get_access_key():
    return os.getenv("ACCESS_KEY")


def get_bohrium_project_id():
    return int(os.getenv("BOHRIUM_PROJECT_ID"))


def submit_job(client: OpenSDK):
    ext_config = {"jobConfig": {"jobRead": 1}}
    inputs = {
        "center_x": 0,
        "center_y": 0,
        "center_z": 0,
        "input_ligands": [
            UploadInputItem(
                src=Path(__file__).parent / "data" / "uni_dock" / "uniDock_ligand.sdf"
            )
        ],
        "input_protein": UploadInputItem(
            src=Path(__file__).parent / "data" / "uni_dock" / "uniDock_protein.pdb"
        ),
        "local_only": False,
        "mode": "fast",
        "num_modes": 1,
        "random_seed": False,
        "score_only": False,
        "scoring": "vinardo",
        "size_x": 25,
        "size_y": 25,
        "size_z": 25,
        "unimol_bias": False,
    }

    sub_model_name = "UniDockModel"
    resp = client.app.job.submit(
        app_key="uni-dock",
        sub_model_name=sub_model_name,
        project_id=get_bohrium_project_id(),
        inputs=inputs,
        ext_config=ext_config,
    )
    return resp


def test_submit_job():
    client = OpenSDK(access_key=get_access_key())
    resp = submit_job(client) or {}
    assert resp.get("code") == 0
    assert glom(resp, "data.jobId") is not None


def test_list_jobs():
    client = OpenSDK(access_key=get_access_key())
    res = client.app.job.list() or {}

    assert res.get("code") == 0
    assert glom(res, "data") is not None
    assert isinstance(glom(res, "data.items"), list)


def get_one_job_from_list(client: OpenSDK):
    jobs_rsp = client.app.job.list() or {}
    assert isinstance(glom(jobs_rsp, "data.items"), list)

    jobs = glom(jobs_rsp, "data.items")
    if len(jobs) == 0:
        assert False, "No job is running or finished"

    # APP_JOB_STATUS_WAIT_SUBMIT = -10 //预创建
    # APP_JOB_STATUS_SUBMIT      = 0  //已提交
    # APP_JOB_STATUS_RUNNING     = 1  // 运行中
    # APP_JOB_STATUS_SUCCESS     = 10 //运行成功
    # APP_JOB_STATUS_FAIL        = 30 //运行失败
    # APP_JOB_STATUS_KILL        = 31 //人工停止
    accessible_jobs = list(filter(lambda x: x.get("status") == 10, jobs))
    if not accessible_jobs:
        accessible_jobs = list(filter(lambda x: x.get("status") == 1, jobs))
    if not accessible_jobs:
        assert False, "No job is running or finished"

    current_job = accessible_jobs[0]
    return current_job


def test_get_job():
    client = OpenSDK(access_key=get_access_key())
    current_job = get_one_job_from_list(client)

    assert type(current_job.get("id")) == int
    assert current_job.get("jobName") is not None

    current_job_id = current_job.get("id")

    job_rsp = client.app.job.detail(current_job_id) or {}
    assert job_rsp.get("code") == 0
    assert glom(job_rsp, "data") is not None

    current_job_data = glom(job_rsp, "data")
    assert current_job_data.get("id") == current_job_id
    assert current_job_data.get("jobName") == current_job.get("jobName")


def test_download_job_result():
    client = OpenSDK(access_key=get_access_key())
    current_job = get_one_job_from_list(client)

    job_id = current_job.get("id")

    import tempfile

    with tempfile.TemporaryDirectory() as temp_dir:
        print(f"Temporary directory created: {temp_dir}")
        temp_dir = Path(temp_dir)
        is_ok = client.app.job.download(
            job_id, remote_target="config.json", save_path=temp_dir / "outputs"
        )

        assert is_ok
        assert (temp_dir / "outputs").exists()
        assert (temp_dir / "outputs" / "config.json").exists()

        is_ok = client.app.job.download(
            job_id, remote_target="task", save_path=temp_dir / "task"
        )

        assert is_ok
        assert (temp_dir / "task").exists()
        assert (temp_dir / "task").is_dir()


def test_get_job_log():
    client = OpenSDK(access_key=get_access_key())
    current_job = get_one_job_from_list(client)

    job_id = current_job.get("id")

    log_rsp = client.app.job.log(job_id) or {}

    assert log_rsp.get("code") == 0
    assert glom(log_rsp, "data") is not None
    assert glom(log_rsp, "data.log") is not None


def test_kill_job():
    client = OpenSDK(access_key=get_access_key())

    resp = submit_job(client) or {}

    assert resp.get("data") is not None
    assert glom(resp, "data.jobId") is not None

    time.sleep(5)

    job_id = glom(resp, "data.jobId")

    kill_rsp = client.app.job.kill(job_id) or {}

    assert kill_rsp.get("code") == 0

    resp = client.app.job.detail(job_id) or {}

    assert resp.get("code") == 0
    assert glom(resp, "data") is not None
    assert glom(resp, "data.status") == 31


def test_delete_job():
    client = OpenSDK(access_key=get_access_key())

    resp = submit_job(client) or {}
    assert resp.get("data") is not None
    assert glom(resp, "data.jobId") is not None

    time.sleep(5)

    job_id = glom(resp, "data.jobId")

    delete_res = client.app.job.delete(job_id)

    assert delete_res.get("code") != 0

    kill_rsp = client.app.job.kill(job_id)
    assert kill_rsp.get("code") == 0
    delete_res = client.app.job.delete(job_id)
    assert delete_res.get("code") == 0



import logging
import os
import uuid
from pathlib import Path
from typing import Optional
from unittest import result

from ..._exceptions import NotFoundError
from ..._resource import SyncAPIResource
from ..._response import APIResponse
from ..._tiefblue_client import Tiefblue
from ...types.job.job import JobAddRequest

logger = logging.getLogger(__name__)


class Job(SyncAPIResource):

    def detail(self, job_id):
        uri = f"/{self._client.api_prefix}/v1/job/{job_id}"
        response = self._client.get(uri)
        return APIResponse(response).json

    def submit(
        self,
        *,
        project_id: int,
        job_name: str,
        machine_type: str,
        cmd: str,
        image_address: str,
        job_group_id: int = 0,
        work_dir: str = "",
        dataset_path: list = [],
        log_files: list = [],
        out_files: list = [],
    ):
        data = self.create_job(project_id, job_name, job_group_id)

        if work_dir != "":
            if not os.path.exists(work_dir):
                raise FileNotFoundError
            if os.path.isdir(work_dir):
                self.upload_dir(work_dir, data.storePath, data.token)
            else:
                file_name = os.path.basename(work_dir)
                object_key = os.path.join(data["storePath"], file_name)
                self.upload(work_dir, object_key, data["token"])

        ep = os.path.expanduser(result)
        p = Path(ep).absolute().resolve()
        p = p.joinpath(str(uuid.uuid4()) + "_temp.zip")

        job_add_request = JobAddRequest(
            download_path=str(p.absolute().resolve()),
            dataset_path=dataset_path,
            job_name=job_name,
            project_id=project_id,
            job_id=data["jobId"],
            oss_path=data["storePath"],
            image_name=image_address,
            scass_type=machine_type,
            cmd=cmd,
            log_files=log_files,
            out_files=out_files,
        )
        return self.insert(job_add_request.to_dict())

    def insert(self, data):
        uri = f"/{self._client.api_prefix}/v2/job/add"
        response = self._client.post(uri, json=data)
        return APIResponse(response).json

    def delete(self, job_id):
        uri = f"/{self._client.api_prefix}/v1/job/del/{job_id}"
        response = self._client.post(uri)
        return APIResponse(response).json

    def terminate(self, job_id):
        uri = f"/{self._client.api_prefix}/v1/job/terminate/{job_id}"
        response = self._client.post(uri)
        return APIResponse(response).json

    def kill(self, job_id):
        uri = f"/{self._client.api_prefix}/v1/job/kill/{job_id}"
        response = self._client.post(uri)
        return APIResponse(response).json

    def log(self, job_id, log_file="STDOUTERR", page=-1, page_size=8192):
        uri = f"/{self._client.api_prefix}/v1/job/{job_id}/log"
        response = self._client.get(
            uri,
            params={"logFile": log_file, "page": page, "pageSize": page_size},
        )
        return APIResponse(response).json

    def create_job(
        self,
        project_id: int,
        name: Optional[str] = None,
        group_id: Optional[int] = 0,
    ):
        uri = f"/{self._client.api_prefix}/v1/job/create"
        data = {
            "projectId": project_id,
            "name": name,
            "bohrGroupId": group_id,
        }
        response = self._client.post(uri, json=data)
        return APIResponse(response).data

    def create_job_group(self, project_id, job_group_name):
        uri = f"/{self._client.api_prefix}/v1/job_group/add"
        response = self._client.post(
            uri,
            json={"name": job_group_name, "projectId": project_id},
        )

        return APIResponse(response).json

    def upload(
        self,
        file_path: str,
        object_key: str,
        token: str,
    ):
        tiefblue = Tiefblue()
        custom_headers = {"Authorization": token}

        tiefblue.upload_from_file_multi_part(
            object_key=object_key,
            custom_headers=custom_headers,
            file_path=file_path,
            progress_bar=True,
        )

    def upload_dir(self, work_dir, store_path, token):
        # TODO: upload rule
        if not work_dir.endswith("/"):
            work_dir = work_dir + "/"
        for root, _, files in os.walk(work_dir):
            for file in files:
                full_path = os.path.join(root, file)
                object_key = full_path.replace(work_dir, store_path)
                self.upload(full_path, object_key, token)

    def download(self, job_id, save_path):
        detail = self.detail(job_id)

        result_url = detail.get("data").get("resultUrl")

        if not result_url:
            raise NotFoundError("Result url not found")

        # check save path
        # if not exist create it
        if not os.path.exists(os.path.dirname(save_path)):
            os.makedirs(os.path.dirname(save_path))

        logger.debug(
            f"Downloading job {job_id} to {save_path}, result url: {result_url}"
        )

        tiefblue = Tiefblue()
        tiefblue.download_from_url(result_url, save_path)



import logging

from ..._resource import SyncAPIResource
from ..._response import APIResponse

log = logging.getLogger(__name__)


class User(SyncAPIResource):

    def getInfo(self):
        response = self._client.get("openapi/v1/ak/get")
        return APIResponse(response).json
